from ordf.graph import Graph
from ordf.handler import init_handler
from ordf.namespace import register_ns, Namespace, RDF, RDFS, FOAF, OWL, SKOS
from ordf.term import Literal
from ordf.vocab.owl import Class

def fuxi_handler(default={}):
    config = {
        "ordf.readers": "fuxi,rdflib",
        "ordf.writers": "fuxi,rdflib",
        }
    config.update(default)
    return init_handler(config)

EX = Namespace("http://example.org/")
register_ns("ex", EX)

def test_01_rdfs():
    h = fuxi_handler({ "fuxi.args": "ordf.vocab.rdfs" })

    g = Graph(identifier=EX["Max"])
    g.add((EX["Max"], RDF["type"], EX["Cat"]))
    h.put(g)
    
    r = h.get(EX["Max"])
    assert r.exists((EX["Max"], RDF["type"], RDFS["Resource"]))

def test_02_owl():
    h = fuxi_handler({ "fuxi.args": "ordf.vocab.owl" })

    ctx = h.context("test", "cat")

    cat = Class(EX["Cat"], graph=Graph(identifier=EX["Cat"]))
    ctx.add(cat.graph)

    cateq = Class(EX["CatEquiv"],
                  equivalentClass=[EX["Cat"]],
                  graph=Graph(identifier=EX["CatEquiv"]))
    ctx.add(cateq.graph)

    max = Graph(identifier=EX["Max"])
    max.add((EX["Max"], RDF["type"], EX["Cat"]))
    ctx.add(max)

    ctx.commit()

    cat = h.get(EX["Cat"])
    assert cat.exists((EX["Cat"], OWL["equivalentClass"], EX["CatEquiv"]))

def test_03_skos():
    h = fuxi_handler({ "fuxi.args": "ordf.vocab.skos" })
    ctx = h.context("test", "cat")

    animal = Class(EX["Animal"], graph=Graph(identifier=EX["Animal"]))
    animal.graph.add((EX["Animal"], RDF["type"], SKOS["Concept"]))
    ctx.add(animal.graph)

    cat = Class(EX["Cat"], graph=Graph(identifier=EX["Cat"]))
    cat.graph.add((EX["Cat"], SKOS["broaderTransitive"], EX["Animal"]))
    ctx.add(cat.graph)

    tabby = Class(EX["Tabby"], graph=Graph(identifier=EX["Tabby"]))
    tabby.graph.add((EX["Tabby"], SKOS["broaderTransitive"], EX["Cat"]))
    ctx.add(tabby.graph)

    ctx.commit()

    animal = h.get(EX["Animal"])
    assert animal.exists((EX["Animal"], SKOS["narrowerTransitive"], EX["Cat"]))
    assert animal.exists((EX["Animal"], SKOS["narrowerTransitive"], EX["Tabby"]))

def test_04_retract():
    print
    h = fuxi_handler({ "fuxi.args": "ordf.vocab.rdfs" })
    ctx = h.context("test", "retract")
    
    max = Graph(identifier=EX["Max"])
    max.add((EX["Max"], RDF["type"], EX["Cat"]))
    ctx.add(max)
    cs = ctx.commit()

    max = h.get(EX["Max"]) ## should contain inferred triples
    assert max.exists((EX["Max"], RDF["type"], RDFS["Resource"]))

    ## empty the rule sets, would not normally be done so
    ## no public API for this, really
    h.fuxi.rule_modules = []
    del h.fuxi.network

    max.add((EX["Max"], FOAF["name"], Literal("Max")))
    ctx = h.context("test", "retract")
    ctx.add(max)
    cs = ctx.commit()

    max = h.get(EX["Max"]) ## should not contain inferred triples
    assert not max.exists((EX["Max"], RDF["type"], RDFS["Resource"]))



from ordf.graph import Graph
from ordf.namespace import register_ns

register_ns("umbel", "http://umbel.org/umbel#")
register_ns("sc", "http://umbel.org/umbel/sc/")
register_ns("ac", "http://umbel.org/umbel/ac/")
register_ns("ne", "http://umbel.org/umbel/ne/")

from ordf.namespace import UMBEL, SC, AC, NE

from logging import getLogger
log = getLogger(__name__)

umbel_n3 = "http://umbel.org/ontology/umbel.n3"
ext_n3 = "http://umbel.org/ontology/umbel_external_ontologies_linkage.n3"

def rdf_data():
    def fetch(src):
        log.info("Fetching %s" % src)
        g = Graph(identifier=src).parse(src, format="n3")
        log.info("Parsed %s (%d triples)" % (src, len(g)))
        return g

    yield fetch(umbel_n3)
    yield fetch(ext_n3)
    yield fetch("http://umbel.org/ontology/umbel_subject_concepts.n3")
    yield fetch("http://umbel.org/ontology/umbel_abstract_concepts.n3")

def inference_rules(handler, network):
    from FuXi.DLP.DLNormalization import NormalFormReduction
    umbel = handler.get(umbel_n3)
    if len(umbel) == 0:
        umbel = Graph(identifier=umbel_n3).parse(umbel_n3, format="n3")
        handler.put(umbel)
    ext = handler.get(ext_n3)
    if len(ext) == 0:
        ext = Graph(identifier=ext_n3).parse(ext_n3, format="n3")
        handler.put(ext)
    onto = Graph()
    onto += umbel
    onto += ext
    NormalFormReduction(onto)
    return network.setupDescriptionLogicProgramming(onto, addPDSemantics=False)



from ordf.namespace import register_ns, Namespace
register_ns("ont", Namespace("http://www.example.org/ontology#"))
register_ns("data", Namespace("http://www.example.org/data#"))
from ordf.graph import Graph
from ordf.namespace import ONT, DATA, SKOS
from ordf.vocab.skos import ConceptScheme, Concept

class Food(ConceptScheme):
    def __init__(self, *av, **kw):
        super(Food, self).__init__(ONT.Food, *av, **kw)

class Fruit(Concept):
    def __init__(self, *av, **kw):
        kwa = kw.copy()
        kwa["skipClassMembership"] = True
        super(Fruit, self).__init__(*av, **kwa)
        if not kw.get("skipClassMembership"):
            self.type = ONT.Fruit

class Veg(Concept):
    def __init__(self, *av, **kw):
        kwa = kw.copy()
        kwa["skipClassMembership"] = True
        super(Veg, self).__init__(*av, **kwa)
        if not kw.get("skipClassMembership"):
            self.type = ONT.Veg

def test_skos():
    ontology = Graph()
    data = Graph()

    food = Food(graph=ontology, factoryGraph=data)

    fruit = Concept(ONT.Fruit, graph=ontology, factoryGraph=data, factoryClass=Fruit)
    fruit.inScheme = food
    food.topConcept = fruit

    veg = Concept(ONT.Veg, graph=ontology, factoryGraph=data, factoryClass=Veg)
    veg.inScheme = food
    food.topConcept = veg

    apple = fruit.get_or_create(DATA.apple)
    apple.broader = fruit

    carrot = veg.get_or_create(DATA.carrot)
    carrot.broader = veg

    assert veg in carrot.broader
    assert fruit in apple.broader

    assert veg in food.topConcept
    assert food in veg.inScheme

    assert fruit in food.topConcept
    assert food in veg.inScheme



import os
import pkg_resources
from getpass import getuser
from logging import getLogger
from ordf.graph import Graph

log = getLogger(__name__)

def rdf_data():
    graph_uri = "http://ordf.org/schema/ordf"
    log.info("Loading %s" % graph_uri)

    graph = Graph(identifier=graph_uri)
    fp = pkg_resources.resource_stream("ordf.vocab", os.path.join("n3", "ordf.n3"))
    graph.parse(fp, format="n3")
    fp.close()

    from ordf.vocab.opmv import Agent, Process
    agent = Agent()
    agent.nick(getuser())
    proc = Process()
    proc.result(graph)

    yield graph



from ordf.handler.queue import RabbitHandler
from ordf.handler import Handler, HandlerPlugin
from ordf.tests import our_test_graph
from time import sleep

class LoggingHandler(HandlerPlugin):
    def put(self, store):
        for ctx in store.contexts():
            print "recv graph", ctx.identifier

receiver = RabbitHandler(hostname="localhost", userid="guest", password="guest", exchange="ordf_test")
receiver.register_writer(LoggingHandler())

sender = Handler()

class TestClassRabbit(object):
    def test_01_connect_sender(self):
        sender.register_writer("rabbit", hostname="localhost", userid="guest", password="guest", exchange="ordf_test")
        sender.__writers__[0].connect()
        sleep(1)
    def test_02_connect_receiver(self):
        receiver.connect(queue="nosetests")
        sleep(1)
    def test_03_send(self):
        ctx = sender.context("test_rabbit", "test_rabbit")
        ctx.add(our_test_graph())
        ctx.commit()
    def test_10_close(self):
        sleep(1)
        receiver.close()



from ordf.graph import Graph
from ordf.namespace import RDFS
import os, pkg_resources

from logging import getLogger
log = getLogger(__name__)

def rdf_data():
    graph_uri = RDFS[""]
    log.info("Fetching %s" % graph_uri)
    g = Graph(identifier=graph_uri).parse(graph_uri)
    log.info("Parsed %s (%d triples)" % (graph_uri, len(g)))
    yield g

def inference_rules(handler, network):
    from FuXi.Horn.HornRules import HornFromN3
    rule_file = pkg_resources.resource_filename("ordf.vocab", os.path.join("n3", "rdfs-rules.n3"))
    rules = HornFromN3(rule_file)
    for rule in rules:
        network.buildNetworkFromClause(rule)

    from FuXi.DLP.DLNormalization import NormalFormReduction
    rdfs = handler.get(RDFS[""])
    if len(rdfs) == 0:
        for rdfs in rdf_data():
            pass
    NormalFormReduction(rdfs)
    dlp = network.setupDescriptionLogicProgramming(rdfs, addPDSemantics=True)

    return list(rules) + dlp



#
# examples from the manual
#

from ordf.namespace import register_ns, Namespace
register_ns("data", Namespace("http://example.org/data#"))
register_ns("ont", Namespace("http://example.org/ontology#"))

## imports required for the example
from ordf.graph import Graph, ConjunctiveGraph
from ordf.namespace import DATA, ONT, FOAF, RDF, RDFS
from ordf.term import URIRef, Literal
from ordf.vocab.owl import Class, AnnotatibleTerms, predicate, object_predicate

class Person(AnnotatibleTerms):
    def __init__(self, *av, **kw):
        super(Person, self).__init__(*av, **kw)
        self.type = FOAF.Person
    name = predicate(FOAF.name)
    homepage = predicate(FOAF.homepage)

def test_00_person():
    ## create a person
    bob = Person(DATA.bob)
    bob.name = "Bob"
    bob.homepage = URIRef("http://example.org/")

    assert FOAF.Person in bob.type
    assert "Bob" in bob.name
    assert "http://example.org/" in bob.homepage

class Country(AnnotatibleTerms):
    def __init__(self, *av, **kw):
        super(Country, self).__init__(*av, **kw)
        self.type = ONT.Country

class City(AnnotatibleTerms):
    def __init__(self, *av, **kw):
        super(City, self).__init__(*av, **kw)
        self.type = ONT.City
    country = object_predicate(ONT.country, Country)

def test_01_obj_pred():
    data = Graph()

    scotland = Country(DATA.Scotland, graph=data)
    scotland.label = "Scotland"
    
    edinburgh = City(DATA.Edinburgh, graph=data)
    edinburgh.label = "Edinburgh"
    edinburgh.country = scotland

    assert scotland in edinburgh.country

class PlaceClass(Class):
    def __init__(self, **kw):
        super(PlaceClass, self).__init__(ONT.Place, **kw)

class CountryClass(Class):
    def __init__(self, **kw):
        super(CountryClass, self).__init__(ONT.Country, **kw)
        self.subClassOf = PlaceClass(graph=self.graph)
        self.factoryClass = Country
        self.label = "Country"

class CityClass(Class):
    def __init__(self, **kw):
        super(CityClass, self).__init__(ONT.City, **kw)
        self.subClassOf = PlaceClass(graph=self.graph)
        self.factoryClass = City
        self.label = "City"

def test_02_class():
    ontology = Graph()
    data = Graph()

    scotland = Country(DATA.scotland, graph=data)
    scotland.label = "Scotland"
    
    edinburgh = City(DATA.edinburgh, graph=data)
    edinburgh.label = "Edinburgh"
    edinburgh.country = scotland

    country = CountryClass(graph=ontology, factoryGraph=data)
    city = CityClass(graph=ontology, factoryGraph=data)

    assert country.get(DATA.scotland) == scotland
    assert city.get(DATA.edinburgh) == edinburgh

def test_03_filter():
    from telescope import v

    ontology = Graph()
    data = Graph()

    country = CountryClass(graph=ontology, factoryGraph=data)
    city = CityClass(graph=ontology, factoryGraph=data)

    scotland = Country(DATA.scotland, graph=data)
    scotland.label = "Scotland"
    russia = Country(DATA.russia, graph=data)
    russia.label = "Russia"

    results = list(country.filter((v.id, RDFS.label, Literal("Russia"))))
    assert len(results) == 1
    assert 'Russia' in results[0].label

def test_04_dl():
    ontology = Graph()
    data = Graph()

    scotland = Country(DATA.scotland, graph=data)
    scotland.label = "Scotland"
    
    edinburgh = City(DATA.edinburgh, graph=data)
    edinburgh.label = "Edinburgh"
    edinburgh.country = scotland

    place = PlaceClass(graph=ontology, factoryGraph=data)
    country = CountryClass(graph=ontology, factoryGraph=data)
    city = CityClass(graph=ontology, factoryGraph=data)

    from FuXi.Rete.RuleStore import SetupRuleStore
    from FuXi.Rete.Util import generateTokenSet

    rule_store, rule_graph, network = SetupRuleStore(makeNetwork=True)
    network.reset(data)
    network.setupDescriptionLogicProgramming(ontology)
    network.feedFactsToAdd(generateTokenSet(data))

    assert place.get(DATA.scotland) == scotland
    assert place.get(DATA.edinburgh) == edinburgh



from ordf.namespace import RDF, DC, FOAF, OPMV
from ordf.graph import Graph, ConjunctiveGraph
from ordf.term import URIRef
from ordf.tests import fixture, identifier, card_graph
from ordf.vocab.foaf import Person

def test_10_bnc():
    g = fixture("marc.n3")
    proc = g.bnc((None, OPMV["wasGeneratedBy"], None))
    assert len(proc) == 10
    

def test_11_bnc():
    g = fixture("marc_bio.n3")
    subj = g.bnc((None, DC["subject"], None))
    assert len(subj) == 18
    

def test_12_replace():
    g = fixture("marc_bio.n3")
    subj = g.bnc((None, DC["subject"], None))
    repl = subj.replace((None, DC["subject"], None), (identifier, None, None))
    statements = list(repl.triples((identifier, None, None)))
    assert len(statements) == 4



from rdflib.plugin import register
try:
    from rdflib.serializer import Serializer
    from rdflib.plugins.serializers.n3 import N3Serializer
except ImportError:
    from rdflib.syntax.serializer import Serializer
    from rdflib.syntax.serializers.N3Serializer import N3Serializer
from ordf.term import BNode
from StringIO import StringIO
import sys

__all__ = ["Serializer", "TriGSerializer"]

class TriGSerializer(N3Serializer):
    def serialize(self, stream, base=None, encoding=None, spacious=None, **args):
        self.reset()
        if stream is None:
                stream = StringIO()
        self.stream = stream
        self.base = base

        if spacious is not None:
            self._spacious = spacious

        self.preprocess()
        self.startDocument()

        if self.store.context_aware:
            contexts = self.store.contexts()
        else:
            contexts = [self.store]
        for g in contexts:
            self.write("\n%s {" % self.store.qname(g.identifier))
            self.depth += 1
            s = N3Serializer(g, parent=self)
            s.serialize(self.stream)
            self.depth -= 1
            self.write("}\n")

        self.endDocument()
        self.write("\n")
      
        if isinstance(self.stream, StringIO): 
            return self.stream.getvalue()

register('trig', Serializer, 'ordf.serializer', 'TriGSerializer')




"""
This is the operational core of ORDF. In the normal course of events, an
application will,

    1. Instantiate a :class:`Handler` object, which is usually a singleton.
    2. Register one or more handler implementations for reading or writing.
    3. Use the :meth:`get` and :meth:`put` methods to respectively retrieve
       and save *Graphs*.
    4. Use :meth:`context` to save *Graphs* within the context of a
       :class:`ordf.vocab.changeset.ChangeSet`.

See the documentation for :class:`HandlerPlugin` for how handler
the implementation works.

Initialising Handlers
---------------------

.. autoclass:: ConfigError
.. autofunction:: init_handler

Reading and Writing Graphs -- :class:`Handler`
----------------------------------------------

.. autoclass:: Handler

Base for Storage and Indices -- :class:`HandlerPlugin`
------------------------------------------------------

.. autoclass:: HandlerPlugin

Making Changesets -- :class:`ChangeContext`
-------------------------------------------

.. autoclass:: ChangeContext
"""
__all__ = ["Handler", "HandlerPlugin", "init_handler", "ConfigError"]

from ordf.vocab.changeset import ChangeSet
from ordf.graph import Graph, _Graph, ConjunctiveGraph, _CGraph
from ordf.namespace import namespaces, CS, RDF
import pkg_resources
import logging

class ChangeContext(object):
    """
    Takes care of constructing changesets and distributes the results

    Not to be instantiated directly. Returned from the :meth:`Handler.context`
    method.
    
    Typical usage:

    .. code-block:: python

        ctx = handler.context(...)
        ctx.add(graph)
        ctx.add(another)
        ctx.commit()

    .. automethod:: add
    .. automethod:: commit
    .. automethod:: rollback
    """
    def __init__(self, handler, *av, **kw):
        self.handler = handler
        self.__av__ = av
        self.__kw__ = kw
        self.graphs = []
        self.rollback()

    def add(self, graph):
        """
        Add an RDFLib Graph instance to this changeset context
        """
        copy = Graph(identifier=graph.identifier, store=self.store)
        for statement in graph.triples((None, None, None)):
            copy.add(statement)
        self.graphs.append(copy)

    def commit(self):
        """
        Commit any pending changes in this context and distribute
        them via the handler.

        This method actually constructs the :class:`ordf.changeset.ChangeSet`
        instance. This *ChangeSet* is initialised with any positional and 
        keyword arguments that were passed to the present class on creation.

        It then iterates over any graphs that have been added with 
        :meth:add and requests the previous version from the
        :meth:`Handler.get`. The differences between the previous and current
        versions are added to the changeset.

        Now, *ChangeSet* in hand, we call the :meth:`Handler.put` for the
        changeset and then for each of the new versions of the graphs in
        turn.
        """
	kw = self.__kw__.copy()
	kw["store"] = self.store
        cs = ChangeSet(*self.__av__, **kw)

        ## we keep track if there is already a changeset in the store
        embedded_changeset = False
        for graph in self.graphs:
            ## check if is changeset.
            ## do not diff changesets
            is_changeset = False
            for s,p,o in graph.triples((graph.identifier, RDF.type, CS.ChangeSet)):
                is_changeset = True
                embedded_changeset = True
                break
            if is_changeset:
                continue

            ## diff normal graphs
            orig = self.handler.get(graph)
            if orig == None:
                orig = Graph(identifier=graph.identifier)
            changes = cs.diff(orig, graph)

        cs.commit()
        if embedded_changeset or cs:
            self.handler.put(self.store)

        return cs

    def rollback(self):
        """
        Roll back any pending changes in this context

        This simply emptying it of graphs that have been previously added using
        :meth:`add`.
        """
        dummy = Graph("IOMemory")
        self.store = dummy.store


class Handler(object):
    """
    Handle reading and writing of RDF Graphs

    Storage back-ends are registered to an instance of this class for reading and
    writing. It distributes read and write (:meth:get and :meth:put) operations over
    these back ends. 

    Both the write case and some more complex read operations are intimately tied 
    to the use of :class:`ordf.changeset.ChangeSet` to which this class is the
    main entry point - it would be unusual, for example, to create a *ChangeSet*
    directly.

    .. code-block:: python

        ## initialise some storage
        null_storage = HandlerPlugin()

        ## create the handler
        handler = Handler()

        ## register the storage
        handler.register_reader(null_storage)
        handler.register_writer(null_storage)

        ## a read operation: retrieve a graph from storage
        handler.get(graph_identifier)
 
        ## a write operation: saving with a change context
        ctx = handler.context("username", "change reason")
        ctx.add(some_graph)
        ctx.commit()

    If using the :func:init_handler function instead of constructing
    the handler by hand (which is recommended) then for each storage
    module will have a "handler" attribute set by that function that
    refers back to the handler.

    Likewise, the handler will have the name of the storage module
    set as an attribute. e.g.::

    .. code-block:: python

        getattr(handler, "xapian")

    will return the instance of the xapian storage module.

    .. automethod:: register_reader
    .. automethod:: register_writer
    .. automethod:: get
    .. automethod:: put
    .. automethod:: remove
    .. automethod:: context
    .. automethod:: changeset
    .. automethod:: history
    .. automethod:: construct
    .. automethod:: query

    """
    def __init__(self, **kw):
        self.__readers__ = []
        self.__writers__ = []
        self.__kw__ = kw
        self.log = logging.getLogger(__name__)
        ordf = pkg_resources.get_distribution("ordf")
        self.log.info("%s initialised ver %s" % (self, ordf.version))

    def __str__(self):
        return self.__class__.__name__ + "(%s/%s)" % (len(self.__readers__), len(self.__writers__))

    def connect(self, *av, **kw):
        """
        Intended for subclasses, anything that must be done to connect the 
        handler to e.g. a message queue should be done here
        """
    def close(self, *av, **kw):
        """
        Intended for subclasses, do any cleanup required to shut down
        """

    def register_reader(self, handler, *av, **kw):
        """
        Register a writer back-end.

        :param handler: either an instance of (a sub-class of) :class:HandlerPlugin or 
            a string. If a string, then uses :meth:HandlerPlugin.find to locate the
            appropriate implementation.

        :param av: positional arguments passed to the handler's constructor when
            *handler* is named with a string.

        :param kw: ditto keyword arguments.
        """
        if isinstance(handler, basestring):
            handler_cls = HandlerPlugin.find(handler)
            handler = handler_cls(*av, **kw)
        self.__readers__.append(handler)
        self.log.info("%s reading from %s" % (self, handler))

    def register_writer(self, handler, *av, **kw):
        """
        Register a writer back-end.

        Parameters are as for :meth:register_reader
        """
        if isinstance(handler, basestring):
            handler_cls = HandlerPlugin.find(handler)
            handler = handler_cls(*av, **kw)
        self.__writers__.append(handler)
        self.log.info("%s writing to %s" % (self, handler))

    def put(self, *av, **kw):
        """
        Iterates over all of the registered write handlers and calls
        their *put* method.

        The *put* methods are called with the given positional and keyword
        arguments.
        """
        for h in self.__writers__:
            if h.put(*av, **kw):
                return

    def get(self, *av, **kw):
        """
        Iterates over all of the registered write handlers and
        calls their *get* method.

        The first back-end to return a non-*None* value wins and this value
        is returned.

        The *get* methods are called wwith the given positional and keyword
        arguments.
        """
        for h in self.__readers__:
            result = h.get(*av, **kw)
            if result is not None:
                result.handler = self
                return result

    def append(self, *av, **kw):
        """
        Append to the given graph
        """
        for h in self.__writers__:
            h.append(*av, **kw)
            
    def remove(self, *av, **kw):
        """
        Remove the graph from all storage and indices
        """
        for h in self.__readers__:
            h.remove(*av, **kw)

    def context(self, user, reason):
        """
        Return an instance of :class:ChangeContext bound to this
        handler

        :param user: The user (presumed already authenticated via whatever 
            mechanism) requesting the change

        :param reason: A short description of the nature of the changes being
           made.
        """
        return ChangeContext(self, user, reason, **self.__kw__)

    def changeset(self, csid, *av, **kw):
        """
        Given a *ChangeSet* identifier, return a changeset
        """
        g = self.get(csid, *av, **kw)
        return ChangeSet(store=g.store, identifier=g.identifier)

    def history(self, identifier):
        """
        return the history of the graph
        
        :param identifier: a *Graph* or identifier
            
        :return: generator of changesets for the given graph, most
           recent first. Values yielded by the generator will be either
           instances of :class:ordf.changeset.ChangeSet or lists in the
           case of multiple parents. This latter is not well tested.
        """
        graph = self.get(identifier)

        def walk_history(csid):
            cs = self.changeset(csid)
            yield cs
            parents = []
            for s,p,csid in cs.triples((csid, CS.precedingChangeSet, None)):
                parents.append(csid)
            if len(parents) == 1:
                for cs in walk_history(parents[0]):
                    yield cs
            if len(parents) > 1:
                yield [walk_history(csid) for csid in parents]

        csid = graph.version()
        if csid is not None:
            for cs in walk_history(csid):
                yield cs

    def construct(self, identifier):
        """
        Construct the requested graph from stored changesets

        :param identifier: a *Graph* or identifier
        """
        history = list(self.history(identifier))
        history.reverse()

        g = Graph(identifier=identifier)
        for cs in history:
            cs.apply(g)
        return g

    def query(self, q):
        """
        Execute a SPARQL query if we have a back-end that supports
        such.

        :param q: the query
        """
        if hasattr(self, "fourstore"):
            return self.fourstore.query(q, initNs=namespaces)
        elif hasattr(self, "rdflib"):
            prefixes = ["PREFIX %s: <%s>" % (k, namespaces[k]) for k in namespaces]
            q = "\n".join(prefixes) + "\n" + q
            return rdflib.query(q)
        else:
            raise AttributeError("No SPARQL Implementations")

class HandlerPlugin(object):
    """
    Instances of this class implement read and/or write operations on storage
    and indices.

    This is an interface specification for subclassing and null implementation.

    The :meth:find class method is used to create handler implementations that may
    live in various modules. To do this it makes use of the *pkg_resources.EntryPoint*
    mechanism. For example, looking at ORDF's *setup.py* you can find the entrypoints
    for the handler implementations that are bundled with this software::
    
        [ordf.handler]
        pairtree=ordf.handler.pt:PairTree
        rdflib=ordf.handler.rdf:RDFLib
        fourstore=ordf.handler.rdf:FourStore
        xapian=ordf.handler.xap:Xapian
        rabbit=ordf.handler.queue:Rabbit

    .. automethod:: connect
    .. automethod:: find
    .. automethod:: get
    .. automethod:: put
    """
    @classmethod
    def find(cls, name):
        """
        Search in the *pkg_resources.EntryPoint* named *[ordf.handler]*
        for a concrete subclass implementing this interface.
            
        :param name: the name of the plugin to find

        If a result is found but is not a subclass of :class:`HandlerPlugin` a
        *ValueError* is raised. If no result is found, an *ImportError* is
        raised.
        """
        for entrypoint in pkg_resources.iter_entry_points(group="ordf.handler"):
            if entrypoint.name == name:
                plugin = entrypoint.load()
                if not issubclass(plugin, cls):
                    raise ValueError("%s is not a subclass of %s" % (plugin, cls))
                return plugin
        raise ImportError("no plugin in [ordf.handler] named %s" % name)

    def connect(self, *av, **kw):
        """
        """

    def put(self, graph):
        """
        Save or index the given conjunctive graph
        """
        if isinstance(graph, _Graph):
            contexts = [graph]
        else:
            contexts = graph.contexts()
        for ctx in contexts:
            self[ctx] = ctx

    def get(self, identifier):
        """
        Retrieve the requested graph

        :param identifier: may be a string or :class:rdflib.term.URIRef or an instance
            of :class:`rdflib.graph.Graph` in which latter case the graph's *identifier*
            is used as  lookup key.
        """
        return self[identifier]

    def append(self, frag):
        """
        Append the fragment to the given graph
        """
        graph = self.handler.get(frag.identifier)
        if graph is not None:
            graph += frag
        else:
            graph = frag
        self.put(graph)
                                                
    def remove(self, identifier):
        """
        Remove the requested graph. Arguments as for :meth:get
        """
        del self[identifier]

    def __getitem__(self, key):
        """
        Implement in concrete subclasses
        """
    def __setitem__(self, key, value):
        """
        Implement in concrete subclasses
        """
    def __delitem__(self, key):
        """
        Implement in concrete subclasses
        """

class ConfigError(Exception):
    """
    Raised on configuration error
    """

def get_args(config, name):
    av = config.get(name + ".args")
    if av:
        av = [av]
    else:
        av = []
    kw = {}
    pfxlen = len(name) + 1
    for k in config.keys():
        if k.startswith(name + ".") and k != name + ".args" and not name.startswith(name + ".connect."):
            kw[k[pfxlen:]] = config[k]
    return av, kw

def init_handler(config):
    """
    Initialise a handler based on the configuration dictionary. Typically
    *config* will be a section of a configuration file parsed with
    :class:`ConfigParser` either directly or accessed via :data:`pylons.config`

    The type of handler to be created, the reading and writing plugins and
    their initialisation arguments can all be specified::

        [app:main]

        ## handler class to use
        ordf.handler = ordf.handler.Handler

        ## if a handler has a connect() method, it is run with arguments 
        ## following. this is used e.g. for ordf.handler.queue.RabbitHandler
        ordf.connect.queue = readerqueue

        ## reader plugins
        ordf.readers = pairtree

        ## writer plugins
        ordf.writers = pairtree,fourstore,xapian,rabbit

        ## arguments for the back-ends
        pairtree.args = /some/where/data/pairtree
        fourstore.args = kbname,soft_limit=-1
        xapian.args = 127.0.0.1:44332
        rabbit.connect.exchange = foo

    Each storage module added to the handler (as a readoer or as a writer)
    will have a "handler" attribute set by this function that refers back
    to the handler. In this way storage/index modules that require access
    to the handler (currently only :class:`~ordf.handler.fuxi.FuXiReasoner`)
    have it.

    Likewise, the handler will have the name of the storage module
    set as an attribute. e.g.::

    .. code-block:: python

        getattr(handler, "xapian")

    will return the instance of the xapian storage module. In this way 
    application code can treat the *handler* instance as a singleton and
    access the various back-ends simply. It is then possible to call 
    specialised search or other methods as needed.
    """

    av, kw = get_args(config, "ordf.handler")
    if "ordf.handler" in config:
        modname, clsname = config["ordf.handler"].split(":")
        mod = __import__(modname, globals(), locals(), [clsname])
        cls = getattr(mod, clsname)
    else:
        cls = Handler
    handler = cls(*av, **kw)

    av, kw = get_args(config, "ordf.connect")
    handler.connect(*av, **kw)

    ## special
    readers = config.get("ordf.readers", None)
    if readers is not None:
        for reader in readers.split(","):
            if not reader: continue
            if hasattr(handler, reader):
                storage = getattr(handler, reader)
            else:
                av, kw = get_args(config, reader)
                storage_cls = HandlerPlugin.find(reader)
                storage = storage_cls(*av, **kw)
                setattr(handler, reader, storage)
                setattr(storage, "handler", handler)
                av, kw = get_args(config, reader + ".connect")
                storage.connect(*av, **kw)
            handler.register_reader(storage)

    writers = config.get("ordf.writers", None)
    if writers is not None:
        for writer in writers.split(","):
            if not writer: continue
            if hasattr(handler, writer):
                storage = getattr(handler, writer)
            else:
                av, kw = get_args(config, writer)
                storage_cls = HandlerPlugin.find(writer)
                storage = storage_cls(*av, **kw)
                setattr(handler, writer, storage)
                setattr(storage, "handler", handler)
                av, kw = get_args(config, writer + ".connect")
                storage.connect(*av, **kw)
            handler.register_writer(storage)

    return handler



from ordf.vocab.opmv import Agent, Process
from ordf.graph import Graph
from ordf.namespace import RDF
from ordf.term import Literal
from pkg_resources import get_distribution

def test_01_proc_version():
    g = Graph()
    a = Agent()
    a.nick("test")
    p = Process()
    p.agent(a)
    p.result(g)

    dist = get_distribution("ordf")
    assert g.exists((None, RDF["value"], Literal(dist.version)))



from ordf.graph import Graph
from ordf.namespace import BIBO
from logging import getLogger

log = getLogger(__name__)

def rdf_data():
    log.info("Fetching %s" % BIBO)
    g = Graph(identifier=BIBO[""]).parse(BIBO)
    log.info("Parsed %s (%d triples)" % (BIBO, len(g)))
    yield g

def inference_rules(handler, network):
    from FuXi.DLP.DLNormalization import NormalFormReduction
    bibo = handler.get(BIBO[""])
    if len(bibo) == 0:
        for bibo in rdf_data():
            handler.put(bibo)
    NormalFormReduction(bibo)
    return network.setupDescriptionLogicProgramming(bibo, addPDSemantics=False)



"""
Namespace Declarations
======================

In addition to the following RDF namespace declarations, this module
conveniently works around some refactoring that `RDFLib`_ has experienced
to make importing the :class:`~rdflib.namespace.Namespace` class more
uniform.

.. autofunction:: bind_ns
.. autofunction:: register_ns

.. data:: namespaces

A dictionary containing all of the namespaces defined here. The keys are the
names of the namespaces and the values are the namespaces themselves.

.. data:: BIBO
.. data:: BIO
.. data:: CS
.. data:: DBPPROP
.. data:: DCES
.. data:: DC
.. data:: DCAM
.. data:: DOAP
.. data:: FOAF
.. data:: FRBR
.. data:: FRESNEL
.. data:: GND
.. data:: GR
.. data:: LIST
.. data:: LOG
.. data:: OPMV
.. data:: ORDF
.. data:: ORE
.. data:: OWL
.. data:: RDAGR2
.. data:: RDARELGR2
.. data:: RELATIONSHIP
.. data:: SCOVO
.. data:: SKOS
.. data:: TRIG
.. data:: VOID
.. data:: WOT
.. data:: UUID

.. _RDFLib: http://www.rdflib.net/
"""

try:
    from rdflib.namespace import Namespace, RDF, RDFS, XSD
except ImportError:
    from rdflib import Namespace
    RDF = Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    RDFS = Namespace('http://www.w3.org/2000/01/rdf-schema#')
    XSD = Namespace('http://www.w3.org/2001/XMLSchema#')

BIBO = Namespace("http://purl.org/ontology/bibo/")
BIO = Namespace("http://purl.org/vocab/bio/0.1/")
CS = Namespace("http://purl.org/vocab/changeset/schema#")
DBPPROP = Namespace("http://dbpedia.org/property/")
DCES = Namespace("http://purl.org/dc/elements/1.1/")
DC = Namespace("http://purl.org/dc/terms/")
DCAM = Namespace("http://purl.org/dc/dcam/")
DOAP = Namespace("http://usefulinc.com/ns/doap#")
FOAF = Namespace("http://xmlns.com/foaf/0.1/")
FRBR = Namespace("http://purl.org/vocab/frbr/core#")
FRESNEL = Namespace("http://www.w3.org/2004/09/fresnel#")
GND = Namespace("http://d-nb.info/gnd/")
GR = Namespace("http://bibliographica.org/schema/graph#")
ICAL = Namespace("http://www.w3.org/2002/12/cal/ical#")
LIST = Namespace("http://www.w3.org/2000/10/swap/list#")
LOG = Namespace("http://www.w3.org/2000/10/swap/log#")
OPMV = Namespace("http://purl.org/net/opmv/ns#")
ORDF = Namespace("http://purl.org/NET/ordf/")
ORE = Namespace("http://www.openarchives.org/ore/terms/")
OWL = Namespace("http://www.w3.org/2002/07/owl#")
RDAGR2 = Namespace("http://RDVocab.info/ElementsGr2/")
RDARELGR2 = Namespace("http://metadataregistry.org/uri/schema/RDARelationshipsGR2/")
RDFG = Namespace("http://www.w3.org/2004/03/trix/rdfg-1/")
RELATIONSHIP = Namespace("http://purl.org/vocab/relationship/")
REV = Namespace("http://purl.org/stuff/rev#")
SCOVO = Namespace("http://purl.org/NET/scovo#")
SIOC = Namespace("http://rdfs.org/sioc/ns#")
SKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
TIME = Namespace("http://www.w3.org/2006/time#")
VOID = Namespace("http://rdfs.org/ns/void#")
WOT = Namespace("http://xmlns.com/wot/0.1/")

UUID = Namespace("urn:uuid:")

namespaces = {
    "rdf": RDF,
    "rdfs": RDFS,
    "bibo": BIBO,
    "bio": BIO,
    "cs": CS,
    "dbpprop": DBPPROP,
    "dces": DCES,
    "dc": DC,
    "dcam": DCAM,
    "doap": DOAP,
    "foaf": FOAF,
    "frbr": FRBR,
    "fresnel": FRESNEL,
    "gnd": GND,
    "gr": GR,
    "ical": ICAL,
    "list": LIST,
    "log": LOG,
    "opmv": OPMV,
    "ordf": ORDF,
    "ore": ORE,
    "owl": OWL,
    "rdaGr2": RDAGR2,
    "rdaRelGr2": RDARELGR2,
    "rdfg": RDFG,
    "relationship": RELATIONSHIP,
    "rev": REV,
    "scovo": SCOVO,
    "sioc": SIOC,
    "skos": SKOS,
    "time": TIME,
    "uuid": UUID,
    "void": VOID,
    "wot": WOT,
    "xsd": XSD
}

def bind_ns(g, namespaces=namespaces):
    """
    Given an :class:`~rdflib.graph.Graph`, bind the namespaces present in
    the dictionary in this module to it for more readable serialisations.

    :param g: an instance of :class:`rdflib.graph.Graph`.
    """
    try:
        [g.bind(*x) for x in namespaces.items()]
    except: ### XXX sometimes gives errors with rdflib sleepycat !?
        pass

def register_ns(prefix, ns):
    """
    Register a namespace for use by ORDF

    :param prefix: the namespace prefix that should be used
    :param namespace: an instance of ordf.namespace.Namespace
    """
    from ordf import namespace
    from logging import getLogger
    log = getLogger("ordf.namespace")
    ns = Namespace(str(ns))
    if hasattr(namespace, prefix.upper()):
        log.warning("Registering %s: <%s> prefix already exists" % (prefix, ns))
    log.debug("Registering %s: <%s>" % (prefix, ns))
    setattr(namespace, prefix.upper().replace("-", "_"), ns)
    namespaces[prefix] = ns

def _init_ns():
    import pkg_resources
    for entrypoint in pkg_resources.iter_entry_points(group="ordf.namespace"):
        init_ns = entrypoint.load()
        init_ns()



__all__ = ["Aggregation"]

from ordf.graph import Graph
from ordf.namespace import RDF, ORE
from ordf.utils import get_identifier

class Aggregation(Graph):
    """
    This flavour of :class:`Graph` exposes aggregations
    as defined by the openarchives ORE vocabulary.
    """
    def __init__(self, *av, **kw):
        super(Aggregation, self).__init__(*av, **kw)
        if not self:
            self.add((self.identifier, RDF["type"], ORE["Aggregation"]))
        else:
            assert self.exists((self.identifier, RDF["type"], ORE["Aggregation"]))

    def aggregate(self, graph):
        self.add((self.identifier, ORE["aggregates"], graph.identifier))
        graph.add((graph.identifier, ORE["isAggregatedBy"], self.identifier))
        
    def contexts(self):
        for s,p,o in self.triples((self.identifier, ORE["aggregates"], None)):
            yield Graph(store=self.store, identifier=o)



"""
Content-Type Negotiation
------------------------

.. data:: graph_serialisations
.. autofunction:: graph_format
.. data:: sparql_serialisations
.. autofunction:: sparql_format

"""

graph_serialisations = {
    "application/rdf+xml" : "pretty-xml",
    "text/n3" : "n3",
    "text/plain" : "nt",
}
def graph_format(request):
    if "format" in request.GET:
        accept = request.GET["format"]
    else:
        accept = request.accept.best_match(graph_serialisations.keys() + ["text/html"])
    return accept, graph_serialisations.get(accept, "html")

sparql_serialisations = {
    "application/sparql-results+json": "json",
}
def sparql_format(request):
    if "format" in request.GET:
        accept = request.GET["format"]
    else:
        accept = request.accept.best_match(sparql_serialisations.keys() + ["text/html"])
    return accept, sparql_serialisations.get(accept, "html")

search_serialisations = {
    "application/javascript": "json"
}
def search_format(request):
    if "format" in request.GET:
        accept = request.GET["format"]
    else:
        accept = request.accept.best_match(search_serialisations.keys() + ["text/html"])
    return accept, search_serialisations.get(accept, "html")



"""
Test Pairtree storage of models
"""
import os
import tempfile
import shutil

from ordf.tests import our_test_graph, identifier
from ordf.handler import Handler
from ordf.handler.pt import PairTree


class TestPairtreeStorage:
    @classmethod
    def setup_class(self):
        self.tmpdir = os.path.join(tempfile.gettempdir(), self.__name__)
        self.ptree = PairTree(self.tmpdir)

    @classmethod
    def teardown_class(self):
        shutil.rmtree(self.tmpdir)

    def test_01_create(self):
        self.ptree[identifier] = our_test_graph()

    def test_02_read(self):
        fsobj = self.ptree[identifier]

    def test_03_delete(self):
        del self.ptree[identifier]

    def test_04_graphops(self):
        self.ptree.put(graph=our_test_graph())
        self.ptree.get(our_test_graph())
        self.ptree.remove(our_test_graph())

    def test_05_handler(self):
        handler = Handler()
        handler.register_reader(self.ptree)
        handler.register_writer(self.ptree)
        ctx = handler.context("ptree_test", "ptree_test")
        ctx.add(our_test_graph())
        ctx.commit()

        for cs in handler.history(identifier):
            self.ptree.remove(cs)
        self.ptree.remove(our_test_graph())




from setuptools import setup, find_packages
import sys, os

version = '0.21'

try:
    from mercurial import ui, hg, error
    repo = hg.repository(ui.ui(), ".")
    ver = repo[version]
except ImportError:
    pass
except error.RepoLookupError:
    tip = repo["tip"]
    version = version + ".%s.%s" % (tip.rev(), tip.hex()[:12])
except error.RepoError:
    pass

setup(
    name='ordf',
    version=version,
    description="Open Knowledge Foundation RDF",
    long_description="""\
Open Knowledge Foundation RDF Library""",
    # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Framework :: Paste",
        "Framework :: Pylons",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.6",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="rdf rdflib provenance messaging",
    author='Open Knowledge Foundation',
    author_email='okfn-help@lists.okfn.org',
    url="http://ordf.org/",
    license='AGPL',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "Paste",
        "setuptools",
        "rdflib",
        #"carrot", carrot is an optional dependency
        "pairtree",
    ],
    entry_points="""
        # -*- Entry points: -*-
        [console_scripts]
        ordf=ordf.command:ordf
        ordf_load=ordf.command:load_rdf
        fresnel=ordf.vocab.fresnel:render

        [ordf.handler]
        null=ordf.handler:HandlerPlugin
        pairtree=ordf.handler.pt:PairTree
        rdflib=ordf.handler.rdf:RDFLib
        fourstore=ordf.handler.rdf:FourStore
        xapian=ordf.handler.xap:Xapian
        redis=ordf.handler.cache:Redis
        rabbit=ordf.handler.queue:RabbitQueue
        fuxi=ordf.handler.fuxi:FuXiReasoner
    """,
)



"""
ControllerFactory
-----------------

.. autofunction:: ControllerFactory
"""

def ControllerFactory(ClassName, TemplateClass, base, handler):
    """
    Factory class for creating a controller using one of the implementations
    in L{ordf.pylons}. This function is generally internal, used by the 
    implementations themselves to define a I{FooControllerFactory} which
    will only take the I{base} and I{handler} arguments. See, for example,
    L{ordf.pylons.graph.GraphControllerFactory}.

    @param ClassName: The name of the class to be returned
    @type ClassName: string

    @param TemplateClass: The implementation of the controller
    @type TemplateClass: a bare class

    @param base: The pylons project's lib.base module
    @type base: module

    @param handler: Message handler supporting read/write operations
    @type handler: instance of L{ordf.handler.Handler}
    """
    @property
    def store(self):
        return model.store
    def render(self, *av, **kw):
        return base.render(*av, **kw)
    return type(ClassName, (TemplateClass, base.BaseController),
                { "handler": handler, "render": render,
                  "__doc__": TemplateClass.__doc__ })



__all__ = ["BNode", "Identifier", "Literal", "Node", "URIRef", "Variable"]

try:
    from rdflib.term import BNode, Identifier, Literal, Node, URIRef, Variable
except ImportError:
    from rdflib import URIRef, BNode, Literal, Variable
    from rdflib.Identifier import Identifier
    from rdflib.Node import Node



"""
This is the core read/write storage module in ORDF. Its purpose is
to store RDF graphs in the filesystem in a specialised directory
hierarchy known as a Pairtree_. The reasoning behind this is explained
in the :ref:`rdf-back-end-design` section of this documentation.

.. autoclass:: PairTree
   :show-inheritance:

.. _Pairtree: https://confluence.ucop.edu/display/Curation/PairTree
"""

from pairtree import PairtreeStorageFactory, PartNotFoundException
from pairtree.pairtree_path import id_to_dirpath
from ordf.graph import Graph, _Graph
from ordf.term import URIRef
from ordf.utils import uuid, get_identifier
from ordf.handler import HandlerPlugin
from traceback import format_exc
from logging import getLogger

log = getLogger(__name__)

class PairTree(HandlerPlugin):
    """
    When adding a *Graph* to this store the first thing that is done is its 
    *uri* is normalised to the *urn:uuid:* namespace. If its idenfifier is
    already in that namespace, nothing is changed. If it is not, it is 
    converted. The original identifier is stored in the filesystem for later
    retrieval so that when a read operation is performed, the result is a
    *Graph* with the correct *uri* identifier. This is done with the utility
    function :func:`ordf.utils.uuid`.
    """
    def __init__(self, store_dir, uri_base="urn:uuid:"):
        f = PairtreeStorageFactory()
        self.store = f.get_store(store_dir=store_dir, uri_base=uri_base)

    def __getitem__(self, key):
        k = uuid(key).lstrip(self.store.uri_base)
        fsobj = self.store.get_object(k)
        #fsobj_dir = id_to_dirpath(k, self.store.pairtree_root)
        #print namaste.get(fsobj_dir)
        try:
            identifier = fsobj.get_bytestream("identifier.txt")
        except PartNotFoundException:
            identifier = get_identifier(key)
        g = Graph(identifier=URIRef(identifier))
        try:
            g.parse(fsobj.get_bytestream("graph.n3", streamable=True), format="n3")
        except PartNotFoundException:
            pass
        return g

    def __setitem__(self, key, g):
        assert isinstance(g, _Graph)
        k = uuid(key).lstrip(self.store.uri_base)
        fsobj = self.store.get_object(k)
        fsobj.add_bytestream("identifier.txt", g.identifier)
        fsobj.add_bytestream("graph.n3", g.serialize(format="n3"))

    def __delitem__(self, key):
        k = uuid(key).lstrip(self.store.uri_base)
        self.store.delete_object(k)

    def __iter__(self):
        for uuid in self.store.list_ids():
            identifier = "urn:uuid:" + uuid
            try:
                graph = self[identifier]
                yield graph
            except GeneratorExit:
                return
            except KeyboardInterrupt:
                raise
            except: 
                log.error("error reading %s:\n%s" % (identifier, format_exc()))



from ordf.tests import our_test_graph
from ordf.rdfobj import RDFObj
from ordf.term import URIRef
from ordf.namespace import *
from pprint import pprint

def test_rdfobj():
    obj = RDFObj(our_test_graph(), URIRef("urn:uuid:db531d1e-93df-46fe-a79f-f0fe4b99d7c4"))
#    pprint (obj)
#    pprint (obj[FRBR.creatorOf])



from ordf.graph import ConjunctiveGraph, Graph
from ordf.vocab.owl import predicate, Ontology, Class, Property, AnnotatibleTerms
from ordf.namespace import register_ns, Namespace
from ordf.term import URIRef, Literal

register_ns("sdmx", Namespace("http://purl.org/linked-data/sdmx#"))
register_ns("sdmxconcept", Namespace("http://purl.org/linked-data/sdmx/2009/concept#"))
register_ns("sdmxdim", Namespace("http://purl.org/linked-data/sdmx/2009/dimension#"))
register_ns("sdmxattr", Namespace("http://purl.org/linked-data/sdmx/2009/attribute#"))
register_ns("sdmxmeasure", Namespace("http://purl.org/linked-data/sdmx/2009/measure#"))
register_ns("sdmxcode", Namespace("http://purl.org/linked-data/sdmx/2009/code#"))
register_ns("sdmxsubject", Namespace("http://purl.org/linked-data/sdmx/2009/subject#"))

from ordf.namespace import SDMX, SDMXDIM, SDMXATTR, VOID, OWL, DC

class DataSet(Class):
    def __init__(self, graph=None):
        super(DataSet, self).__init__(
            SDMX.DataSet,
            subClassOf = [VOID.DataSet],
            graph=graph)
    def get(self, identifier, graph=None):
        if graph is None: graph = self.graph
        ds = _DataSet(identifier, graph=graph)
        ds.type = self.identifier
        ds.factoryGraph = ds.graph
        return ds

class _DataSet(AnnotatibleTerms):
    source = predicate(DC.source)
    isPartOf = predicate(DC.isPartOf)
    structure = predicate(SDMX.structure)
    series = predicate(SDMX.slice)
    
class DataStructureDefinition(Class):
    def __init__(self, graph=None):
        super(DataStructureDefinition, self).__init__(
            SDMX.DataStructureDefinition,
            graph=graph)
    def get(self, identifier=None, graph=None):
        if graph is None: graph = self.graph
        dsd = _DataStructureDefinition(identifier=identifier, graph=graph)
        dsd.type = self.identifier
        dsd.factoryGraph = graph
        return dsd

class _DataStructureDefinition(AnnotatibleTerms):
    component = predicate(SDMX.component)
    
class TimeSeries(Class):
    def __init__(self, graph=None):
        super(TimeSeries, self).__init__(SDMX.TimeSeries, graph=graph)
        self.label = Literal("time series", lang="en")
    def get(self, identifier=None, graph=None):
        if graph is None: graph=self.graph
        series = _Series(identifier=identifier, graph=graph)
        series.type = self.identifier
        series.factoryGraph = graph
        return series

class _Series(AnnotatibleTerms):
    def addDimension(self, dimension, value):
        self.graph.add((self.identifier, dimension, value))
    observation = predicate(SDMX.observation)
    dataSet = predicate(SDMX.dataSet)
    
class Observation(Class):
    def __init__(self, graph=None):
        super(Observation, self).__init__(SDMX.Observation, graph=graph)
        self.label = Literal("observed datum", lang="en")
    def get(self, identifier=None, graph=None):
        if graph is None: graph=self.graph
        obs = _Datum(identifier=identifier, graph=graph)
        obs.type = self.identifier
        obs.factoryGraph = graph
        return obs

class _Datum(AnnotatibleTerms):
    """
    An individual observation, obtained via :meth:`Observation.get`
    """
    def addValue(self, value, measure):
        self.graph.add((self.identifier, SDMX.obsValue, value))
        self.graph.add((self.identifier, SDMXATTR.unitMeasure, measure))
    def addDimension(self, dimension, value):
        self.graph.add((self.identifier, dimension, value))

class Dimension(Property):
    def __init__(self, graph):
        super(Dimension, self).__init__(SDMX.DimensionProperty, graph=graph)
        self.label = Literal("statistical dimension", lang="en")
        self.domain = TimeSeries(graph) | Observation(graph)

def rdf_data():
    graph = Graph(identifier=SDMX[""])
    ds = DataSet(graph)
    ts = TimeSeries(graph)
    obs = Observation(graph)
    dim = Dimension(graph)

    dataset = Property(SDMX["dataSet"], 
                       graph=graph)
    dataset.label = Literal("dataset", lang="en")
    dataset.domain = ts
    dataset.range = ds

    obsValue = Property(SDMX["obsValue"],
                        baseType = OWL.ObjectProperty,
                        graph=graph)
    obsValue.domain = obs

    yield graph

if __name__ == '__main__':
    for g in rdf_data():
        print g.serialize(format="n3")



"""
The ChangeSet Class
-------------------

.. autoclass:: ChangeSet
   :show-inheritance:
"""

__all__ = ["ChangeSet"]

import logging

from ordf.namespace import *
from ordf.term import *
from ordf.utils import *
from ordf.graph import Graph
from ordf.vocab.foaf import Agent
from datetime import datetime
from uuid import uuid1

class ChangeSet(Graph):
    """
    ChangeSet Graph. Typically one does something like,

    .. code-block:: python

        cs = ChangeSet("some name", "some reason")
        cs.diff(g1_orig, g1_new)
        cs.diff(g2_orig, g2_new)
        cs.commit()


    There are two instantiation paths. The usual one where *name* and *reason
    parameters are supplied is intended for constructing new changesets. The
    other where *store* and *identifier* are supplied is intended for accessing
    previously stored changesets.

    :param name: The name of the person or entity creating this changeset. This
                 may be a string or an rdflib datatype (e.g. 
                 :class:`~rdflib.term.URIRef`, :class:`~rdflib.term.Literal`)
    :param reason: A description of the change. This may be a string or an
                   instance of :class:`~rdflib.term.Literal`
    :param store: When obtaining an existing changeset, the :class:`rdflib.store.Store`
                  which contains it.
    :param identifier: When obtaining an existing changeset, the graph identifier
                       that should be used to find it in the store
    :param namespace: When changesets are created they are assigned a name. The
                      name is generated using the :func:`uuid.uuid1` function.
                      It is then appended to the provided namespace.

    .. data:: metadata

        A representation of the metadata of this changeset graph. This excludes
        any *cs:addition* and *cs:removal* properties

    .. automethod:: diff
    .. automethod:: commit
    .. automethod:: rollback
    .. automethod:: apply
    .. automethod:: undo
    """
    def __init__(self, name=None, reason=None, store="IOMemory", identifier=None, namespace=UUID):
        if identifier is None:
            if not isinstance(namespace, Namespace):
                namespace = Namespace(namespace)
            identifier = namespace[unicode(uuid1())]
            self.__frozen__ = False
        else:
            self.__frozen__ = True
        super(ChangeSet, self).__init__(store=store, identifier=identifier)
        bind_ns(self)

        self.log = logging.getLogger("ordf.changeset")
        if name is not None:
            if not isinstance(name, Node):
                name = Literal(name)
                self.agent = Agent()
                self.agent.name(name)
            else:
                self.agent = name
        else:
            self.agent = None

        if reason is not None:
            if isinstance(reason, Node):
                self.reason = reason
            else:
                self.reason = Literal(reason)
        else:
            self.reason = None

        self.changes = 0

    def diff(self, orig, new):
        """
        Populate the ChangeSet with the differences between orig and new.

        :param orig: original graph
        :param new: new graph
        :return: number of distinct changes (additions + removals)
        """
        if self.__frozen__:
            raise Exception("%s already frozen" % self.identifier)
        if self.agent is None:
            raise ValueError("name must not be None")
        if self.reason is None:
            raise ValueError("reason must not be None")
        if orig.identifier != new.identifier:
            raise ValueError("can only diff versions of the same graph")

        ## remove any changeset in the new graph
        new.remove((new.identifier, ORDF.changeSet, None))

        changes = 0

        preceding = []
        ## any triples that have been removed
        for statement in orig.triples((None, None, None)):
            if statement not in new.triples(statement):
                s,p,o = statement
                if s == orig.identifier and p == ORDF.changeSet:
                    preceding.append(o)
                    continue
                self.reify(CS.removal, (s,p,o,orig.identifier))
                changes += 1

        ## any triples that have been added
        for statement in new.triples((None, None, None)):
            if statement not in orig.triples(statement):
                s,p,o = statement
                self.reify(CS.addition, (s,p,o,orig.identifier))
                changes += 1

        if changes > 0:
            ## remove preceding changesets
            for cs in preceding:
                self.reify(CS.removal, (orig.identifier, ORDF.changeSet, cs, orig.identifier))
            ## add linkage to the preceding changeset(s)
            for s,p,o in orig.triples((orig.identifier, ORDF.changeSet, None)):
                self.add((self.identifier, CS.precedingChangeSet, o))

            ## add linkage to the new changeset
            self.add((self.identifier, CS.subjectOfChange, new.identifier))
            add = BNode()
            self.add((self.identifier, CS.addition, add))
            self.add((add, RDF.type, RDF.Statement))
            self.add((add, RDF.subject, new.identifier))
            self.add((add, RDF.predicate, ORDF.changeSet))
            self.add((add, RDF.object, self.identifier))
            self.add((add, ORDF.graph, new.identifier))
  
            ## add linkage to the graph
            new.add((new.identifier, ORDF.changeSet, self.identifier))

        self.changes += changes

        return changes

    def commit(self):
        """
        Commit the changes, mark the changeset read-only.
        """
        if self.__frozen__:
            raise Exception("%s already frozen" % self.identifier)
        self.__frozen__ = True

        if self:
            ## Add changeset metadata
            self.add((self.identifier, RDF.type, CS.ChangeSet))
            self.add((self.identifier, CS.createdDate, Literal(datetime.utcnow().isoformat() + "Z")))
            self.add((self.identifier, DC.creator, self.agent.identifier))
            self += self.agent
            for name in self.agent.objects(self.agent.identifier, FOAF.name):
                self.add((self.identifier, CS.creatorName, name))
            if not self.exists((self.identifier, CS.creatorName, None)):
                for nick in self.agent.objects(self.agent.identifier, FOAF.nick):
                    self.add((self.identifier, CS.creatorName, nick))
            self.add((self.identifier, CS.changeReason, self.reason))
            self.add((self.identifier, ORDF.lens, URIRef("http://ordf.org/lens/changeset")))

            self.log.info("%d changes %s" % (self.changes, self.identifier))
            for s,p,o in self.triples((self.identifier, CS.precedingChangeSet, None)):
                self.log.info("%s follows %s" % (self.identifier, o))

    def rollback(self):
        """
        Empty the changeset. Fails if :meth:commit has already been called.
        """
        if self.__frozen__:
            raise Exception("%s already frozen" % self.identifier)
        self.remove((None, None, None))
        self.changes = 0

    def reify(self, op, (s,p,o,c)):
        add = BNode()
        self.add((self.identifier, op, add))
        self.add((add, RDF.type, RDF.Statement))
        self.add((add, RDF.subject, s))
        self.add((add, RDF.predicate, p))
        self.add((add, RDF.object, o))
        self.add((add, ORDF.graph, c))
        
    def disembody(self, node):
        d = {}
        for s,p,o in self.triples((node, None, None)):
            d[p] = o
        return (d[RDF.subject], d[RDF.predicate], d[RDF.object]), d[ORDF.graph]

    def __cached__(self, attr, predicate):
        if not hasattr(self, attr):
            result = None
            for s,p,o in self.triples((self.identifier, predicate, None)):
                result = o.toPython()
            if result:
                setattr(self, attr, result)
        else:
            result = getattr(self, attr)
        return result
    @property
    def createdDate(self):
        return self.__cached__("__createdDate__", CS.createdDate)

    @property
    def metadata(self):
        m = Graph(identifier=self.identifier)
        bind_ns(m)
        add = 0
        rem = 0
        for s,p,o in self.triples((self.identifier, None, None)):
            if p == CS.addition: add += 1
            elif p == CS.removal: rem += 1
            else: m.add((s,p,o))
        m.add((self.identifier, CS.additionCount, Literal(add)))
        m.add((self.identifier, CS.removalCount, Literal(rem)))
        return m

    def __nonzero__(self):
        nz = False
        ## check for additions, other than changeset housekeeping
        for s,p,o in self.triples((self.identifier, CS.addition, None)):
            (ss, pp, oo), context = self.disembody(o)
            if pp == ORDF.changeSet and oo == self.identifier: ## always add our own ident
                continue
            nz = True
            break
        if not nz:
            ## check for removals, other than changeset housekeeping
            for s,p,o in self.triples((self.identifier, CS.removal, None)):
                (ss, pp, oo), context = self.disembody(o)
                if pp == ORDF.changeSet: ## always remove previous changesets
                    continue
                nz = True
                break
        ## cache the result if we have been committed
        if self.__frozen__:
            self.__nonzero__ = lambda: nz
        return nz

    def _changes(self, graph):
        add, rem = 0, 0
        for s,p,o in self.triples((self.identifier, CS.removal, None)):
            statement, g = self.disembody(o)
            if g == graph.identifier:
                rem += 1
                yield statement, CS.removal
        for s,p,o in self.triples((self.identifier, CS.addition, None)):
            statement, g = self.disembody(o)
            if g == graph.identifier:
                add += 1
                yield statement, CS.addition
        self.log.debug("changes for %s: +%s/-%s" % (graph.identifier, add, rem))

    def apply(self, graph):
        """
        Apply the changeset to a graph
        """
        for (s, p, o), op in self._changes(graph):
            if op == CS.removal:
                graph.remove((s,p,o))
            else:
                graph.add((s,p,o))

    def undo(self, orig):
        """
        Undo the changes in the changeset on a graph.
        """
        for (s, p, o), op in self._changes(orig):
            if op == CS.removal:
                orig.add((s,p,o))
            else:
                orig.remove((s,p,o))



"""
.. autoclass:: Xapian
"""
from ordf.handler import HandlerPlugin
import pkg_resources
import xapian

class Xapian(HandlerPlugin):
    """
    When adding/replacing a graph in the database we use an application 
    specific function to actually create the indexed document. This will
    look in entrypoints of the form::

        [ordf.xapian]
        index = myapp.lib.xapindex:index_graph
        search = myapp.lib.xapindex:search

    :meth:`index_graph` takes a graph and returns a *(docid,
    :class:`xapian.Document`)* tuple.

    For searching, we use another application specific function,
    :meth:search which takes a xapian database connector, query string,
    limit and offset arguments. The result format should be the as 
    returned from :meth:`xapian.Enquire.get_mset`.

    In order to support more than one indexer this class may be passed a
    key-word argument "name" in the constructor to override the key in
    the configuration section.

    .. automethod:: open
    """
    def __init__(self, server, name="index"):
        self.host, port = server.split(":")
        self.port = int(port)
        self.name = name
        self.db = None

    def open(self, writeable=False):
        """
        Slightly complicated routine that allows us to open a connection to
        a xapian database running on a TCP port and cache the handle fo 
        future use. Default is to open read-only.
        """
        if isinstance(self.db, xapian.Database):
            ## we have an open database. tree possibilities
            ## 1. we need a writeable one and we have a read-only one
            if writeable and not isinstance(self.db, xapian.WritableDatabase):
                self.db.close()
            ## 2. we have a writeable one and we need a read-only one
            elif not writeable and isinstance(self.db, xapian.WritableDatabase):
                self.db.close()
            ## 3. we have just the right kind of database, check that it is
            ##    still good
            else:
                try:
                    self.db.reopen()
                    return self.db
                except xapian.NetworkError:
                    pass
        ## either we have no database at this point or we have just closed
        ## the one we had which turned out to be the wrong type
        if writeable:
            self.db = xapian.remote_open_writable(self.host, self.port)
        else:
            self.db = xapian.remote_open(self.host, self.port)
        return self.db

    def close(self):
        if isinstance(self.db, xapian.Database):
            self.db.close()
            self.db = None

    def put(self, store):
        """
        """
        for entrypoint in pkg_resources.iter_entry_points(group="ordf.xapian"):
            if entrypoint.name == self.name:
                index = entrypoint.load()
                db = self.open(writeable=True)
                for docid, doc in index(store):
                    db.replace_document(docid, doc)
                db.flush()
                return
        raise ImportError("no plugin in [ordf.xapian] named %s" % self.name)
